// adjpunct.cpp : R\[ AvP[ṼGg |Cg`܂B
//
/* TeX \[Xt@C@Ǔ_KtB^
   , . -> CB@ɕύX
*/

#include "stdafx.h"

void print_usage( void );
void command_line_info( int argc, _TCHAR* argv[] );
int is_crlf_file( FILE* fp );
int IsKanji1( unsigned char c );
int is_math_env( const char* str );
int is_cs_delimiter( char c );
int check_replace_char( char c, FILE* fp );

const int MAX_PATH = 255;
const int COM_BUF_SIZE = 2048;
static char command[COM_BUF_SIZE];

static char filename[MAX_PATH];
static char output[MAX_PATH];
static FILE* fp_in;
static FILE* fp_out;

static int trans_mode = 1;
static int math_env = 0;

typedef void (*PCF)(FILE* fp_in, FILE* fp_out);
typedef struct tag_spec{
	unsigned char c;
	PCF  cf;
} spec;

PCF is_spec_char( unsigned char c );
int is_replace_char( unsigned char c, FILE* fp_in, FILE* fp_out );
int is_cs_math_begin( char c );
int is_cs_math_end( char c );
int is_cs_replacing_arg( const char* str );
int std_read( FILE* fp_in, FILE* fp_out );
void math_read( FILE* fp_in, FILE* fp_out );
void cs_read( FILE* fp_in, FILE* fp_out );
void space_read( FILE* fp_in, FILE* fp_out );
int env_read( FILE* fp_in, FILE* fp_out );
void arg_read( FILE* fp_in, FILE* fp_out );
void replace_char( FILE* fp_in, FILE* fp_out, int type );
void put_comma( FILE* fp_in, FILE* fp_out );
void put_period( FILE* fp_in, FILE* fp_out );
void skip_spaces( FILE* fp_in );
int is_next_alphabet( FILE* fp_in );

int _tmain(int argc, _TCHAR* argv[])
{
	if( argc == 1 )
	{
		print_usage();
		return 1;
	}
	filename[0] = output[0] = '\0';
	command_line_info( argc, argv );
	if( filename[0] == '\0' )
	{
		fprintf( stderr, "Filename unspecified.\n" );
		return 1;
	}
	FILE* fp = fopen(filename, "rb");
	if( fp == NULL )
	{
		fprintf( stderr, "file %s not found.\n", filename );
		return 1;
	}
	int text_mode = is_crlf_file( fp );
	fclose( fp );
	
	if( (fp_in = fopen(filename, (text_mode)?"rt":"rb")) == NULL )
	{
		fprintf( stderr, "file %s not found.\n", filename );
		return 1;
	}
	fp_out = (output[0] == '\0')?stdout:fopen( output, "wt" );
	if( fp_out == NULL )
	{
		fprintf( stderr, "cannot open %s.\n", output );
		return 1;
	}

	while( std_read(fp_in, fp_out) )
		;

	fclose( fp_in );
	if( output[0] != '\0' ) fclose( fp_out );
	return 0;
}

/*-------------------------------------------------------------------*/
void replace_char( FILE* fp_in, FILE* fp_out, int type )
{
	/* type == 1: comma
	   type == 2: period */
	switch( type )
	{
		case 1:
			put_comma( fp_in, fp_out );
			break;
		case 2:
			put_period( fp_in, fp_out );
			break;
		default:
			break;
	}
	return;
}

/*-------------------------------------------------------------------*/
void put_comma( FILE* fp_in, FILE* fp_out )
{
	fputc( 0x81, fp_out );
	switch( trans_mode )
	{
		case 1:
		case 2: /* C*/
			fputc( 0x43, fp_out );
			break;			
		case 3:
		case 4: /* A*/
			fputc( 0x41, fp_out );
			break;
		default:
			break;
	}
	skip_spaces( fp_in );
	return;
}

/*-------------------------------------------------------------------*/
void put_period( FILE* fp_in, FILE* fp_out )
{
	fputc( 0x81, fp_out );
	switch( trans_mode )
	{
		case 2:
		case 4: /* D*/
			fputc( 0x44, fp_out );
			break;			
		case 1:
		case 3: /* B*/
			fputc( 0x42, fp_out );
			break;
		default:
			break;
	}
	skip_spaces( fp_in );
	return;
}

/**********************************************************************
***                        ǂݍݕ                      ***
**********************************************************************/
int std_read( FILE* fp_in, FILE* fp_out )
{
	int c = fgetc( fp_in );
	if( c == EOF ) return 0;
	PCF pfc;
	int type;
	if( (pfc = is_spec_char(c)) != NULL ) pfc( fp_in, fp_out );
	else if( !math_env )
	{
		/* note that is_replace_char() will put the char unless it must be replaced. */
		if( (type = is_replace_char(c, fp_in, fp_out)) ) replace_char(fp_in, fp_out, type);
	}
	else fputc(c, fp_out);
	return 1;
}

/*-------------------------------------------------------------------*/
void math_read( FILE* fp_in, FILE* fp_out )
{
	/* called just after read $ symbol. */
	int c;
	int org_disp = 0;

	fputc( '$', fp_out );
	if( (c = fgetc(fp_in)) == '$' )
	{
		org_disp = 1;
	}
	if( c == EOF ) return;
	fputc(c,fp_out);
	while( (c = fgetc(fp_in)) != '$' && c != EOF ) fputc(c, fp_out);
	if( c == EOF ) return;
	fputc( c, fp_out );
	if( org_disp )
	{
		fputc( fgetc(fp_in), fp_out ); // must be '$'
	}
	return;
}

/*-------------------------------------------------------------------*/
void cs_read( FILE* fp_in, FILE* fp_out )
{
	/* called just after read \ symbol */
	memset( command, 0, COM_BUF_SIZE * sizeof(char) );
	fputc( '\\', fp_out );
	
	/* \foo ǂ݁Acommand  "foo" B
		\@ ̂悤Ȃ̂͂̎_ŏIB
		fp_in  \foo ̏I_wB
	*/
	
	int c;
	int i = 0;
	long p;
	p = ftell( fp_in );
	while( !is_cs_delimiter( c = fgetc(fp_in) ) && c != EOF )
	{
		command[i++] = c;
		p = ftell( fp_in );
	}
	if( c == EOF )
	{
		fprintf( fp_out, "%s", command );
		return;
	}
	if( i == 0 ) // control char
	{
		fputc( c, fp_out );
		if( is_cs_math_begin(c) ) math_env = 1;
		else if( is_cs_math_end(c) ) math_env = 0;
		return;
	}
	fseek( fp_in, p, SEEK_SET );
	if( !strcmp(command, "begin") )
	{
		if( env_read(fp_in, fp_out) == 1 ) // math env
			math_env = 1;
		return;
	}
	else if( !strcmp(command, "end") )
	{
		if( env_read( fp_in, fp_out ) == 1 ) // math env
			math_env = 0;
		return;
	}
	/* ̒guʂȖߌQ */
	else if( is_cs_replacing_arg(command) )
	{
		fprintf( fp_out, "%s", command );
		return;
	}
	/* math_env == 1 Ȃ疽߂͉̈͂ȂB
	   ͂ȂA" \in [0,1)"̂悤Ȃ̂IȂ[vɂ͂邩 */
	else if( math_env )
	{
		fprintf( fp_out, "%s", command );
		return;
	}
	/* ȉ͈Ǝv̂ToăXLbv */
	else
	{
		arg_read( fp_in, fp_out );
		return;
	}
	return;
}
/*-------------------------------------------------------------------*/
int env_read( FILE* fp_in, FILE* fp_out )
{
	/*
	ǂݎB
	\begin{aaa} Ȃ fp_in  { wB
	\begin {aaa} Ȃ fp_in ͋󔒕wB
	*/
	int c;
	int l = static_cast<int>(strlen(command));
	int i = 0;
	char env_name[256];
	memset( env_name, 0, 256*sizeof(char) );
	
	while( (c = fgetc(fp_in)) != '{' && c != EOF )
	{
		command[l++] = c;
	}
	if( c == EOF )
	{
		fprintf( fp_out, "%s", command );
		return 0;
	}
	command[l++] = '{';
	while( (c = fgetc(fp_in)) != '}' && c != EOF )
	{
		command[l++] = env_name[i++] = c;
	}
	if( c == EOF )
	{
		fprintf( fp_out, "%s", command );
		return 0;
	}
	command[l++] = '}';
	fprintf( fp_out, "%s", command );
	return is_math_env(env_name);
}

/*-------------------------------------------------------------------*/
void arg_read( FILE* fp_in, FILE* fp_out )
{
	/* ߂̈ǂݎB
		\foo {aaa} Ȃ fp_in ͍ŏ̋
		\foo[aaa]{bbb} Ȃ fp_in  [ 
		"\foo "
		"\foo\bar"
		"\foo1"
		̂悤ȏꍇɂɔł
		
		ASY͎̒ʂB
		1. 󔒂ǂݔ΂B
		2. ݂̕B
			2.1 {  [ ȂA}  ] ܂œǂݔ΂B
			2.2 ȊOȂ炽܂Ă镶o͂ĔB
		3. ċAIɎĂԁBiċAj
	*/
	
	int l = static_cast<int>(strlen(command));
	int c;
	int p;

	/* 󔒂ǂݔ΂ */
	while( (p = ftell(fp_in)) && (c = fgetc(fp_in)) == ' ' && c != EOF ) command[l++] = c;
	if( c == EOF )
	{
		fprintf( fp_out, "%s", command );
		return;
	}
	/* c ͋󔒂łȂŏ̕ */
	int n = 1;
	switch( c )
	{
		case '[':
			command[l++] = c;
			while( n != 0 )
			{
				p = ftell( fp_in );
				c = fgetc(fp_in);
				if( c == '[' ) n++;
				else if( c == ']' ) n--;
				command[l++] = c;
			}
			break;
		case '{':
			command[l++] = c;
			while( n != 0 )
			{
				p = ftell( fp_in );
				c = fgetc(fp_in);
				if( c == '{' ) n++;
				else if( c == '}' ) n--;
				command[l++] = c;
			}
			break;
		default:
			fseek( fp_in, p, SEEK_SET ); //Ȃ
			fprintf( fp_out, command );
			return;
			break;
	}
	arg_read( fp_in, fp_out );	
}

/*-------------------------------------------------------------------*/
void space_read( FILE* fp_in, FILE* fp_out )
{
	/* ɋ󔒕ɏoꍇ
	   󔒂ɑǓ_ނłꍇɂ͋󔒂 */
	int num_spaces = 1; // ł1ĂB
	int c;
	long l;
	while( l = ftell(fp_in), (c = fgetc(fp_in)) == ' ' )
		num_spaces++;
	if( !check_replace_char(c, fp_in) )
	{
		for( int i = 0; i < num_spaces; i++ ) fputc( ' ', fp_out );
	}
	fseek( fp_in, l, SEEK_SET );
	return;
}

/*-------------------------------------------------------------------*/
void skip_spaces( FILE* fp_in )
{
	int c;
	long l;
	do{
		l = ftell(fp_in);
	} while( (c = fgetc(fp_in)) == ' ' || c == '\t' );
	fseek( fp_in, l, SEEK_SET );
	return;
}

/**********************************************************************
***                         [eBeB֐                      ***
**********************************************************************/
int is_cs_delimiter( char c )
{
	char cs_delimiter[] = {
		' ', '\\', '0', '1', '2', '3', '4', '5', '6', 
		'7', '8', '9', '!', '"', '#', '$', '%', '&', 
		'\'', '(', ')', '*', '+', ',', '-', '.', '/', 
		':', ';', '=', '?', '@', '[', ']', '^', '`', 
		'~', '|', '{', '[', '}', ']', '<', '>', '_', 
		'@', '\0'
	};
	
	int i = 0;
	while( cs_delimiter[i] != '\0' )
	{
		if( c == cs_delimiter[i++] ) return 1;
	}
	return 0;
}
/*-------------------------------------------------------------------*/
static char* math_env_name[] = {
	"equation",
	"equation*",
	"eqnarray",
	"eqnarray*",
	"math",
	"math*",
	"displaymath",
	"displaymath*",
	"multline",
	"multline*",
	"gather",
	"gather*",
	"align",
	"align*",
	"alignat",
	"alignat*",
	"flalign",
	"flalign*",
	NULL
};
static char cc_math_begin[] = {
	'[',
	'(',
	'\0'
};
static char cc_math_end[] = {
	']',
	')',
	'\0'
};

int is_math_env( const char* str )
{
	int i = 0;
	while( math_env_name[i] != NULL )
	{
		if( !strcmp(str, math_env_name[i++]) ) return 1;
	}
	return 0;
}

/*-------------------------------------------------------------------*/
int is_cs_math_begin( char c )
{
	int i = 0;
	while( cc_math_begin[i] != '\0' )
	{
		if( cc_math_begin[i++] == c ) return 1;
	}
	return 0;
}

/*-------------------------------------------------------------------*/
int is_cs_math_end( char c )
{
	int i = 0;
	while( cc_math_end[i] != '\0' )
	{
		if( cc_math_end[i++] == c ) return 1;
	}
	return 0;
}

/*-------------------------------------------------------------------*/
static char* cs_replacing_arg[] = {
	"mbox",
	"hbox",
	"vbox",
	"pbox",
	"textmc",
	"textgt",
	"textbf",
	"textsf",
	"texttt",
	"textmd",
	"textit",
	"textsl",
	"textsc",
	NULL
};
int is_cs_replacing_arg( const char* str )
{
	int i = 0;
	while( cs_replacing_arg[i] != NULL )
	{
		if( !strcmp(cs_replacing_arg[i++], str) ) return 1;
	}
	return 0;
}

/*-------------------------------------------------------------------*/
static spec spec_table[] = {
	{ '\\', cs_read },
	{ '$',  math_read },
	{ ' ',  space_read },
	{ '\0', NULL }
};
PCF is_spec_char( unsigned char c )
{
	for( int i = 0; ; i++ )
	{
		if( spec_table[i].c == '\0' ) return NULL;
		if( c == spec_table[i].c ) return spec_table[i].cf;
	}
	return NULL;
}

/*-------------------------------------------------------------------*/
int is_crlf_file( FILE* fp )
{
	int c;
	while( (c = fgetc(fp)) != EOF )
	{
		if( c == '\r' ) return 1;
		if( c == '\n' ) return 0;
	}
	return 0;
}

/*-------------------------------------------------------------------*/
int IsKanji1( unsigned char c )
{
	if( (c >= 0x81 && c <= 0x9F) || (c >= 0xE0 && c <= 0xFC) )
		return 1;
	else
		return 0;
}

/*-------------------------------------------------------------------*/
int is_replace_char( unsigned char c, FILE* fp_in, FILE* fp_out )
{
	/*
		A: 8141
		B: 8142
		C: 8143
		D: 8144
	*/
	static int after_alphabet = 0;

	if( c == ',' && !(after_alphabet && is_next_alphabet(fp_in)) ) return 1;
	if( c == '.' && !(after_alphabet && is_next_alphabet(fp_in)) ) return 2;

	if( IsKanji1(c) )
	{
		after_alphabet = 0;
		int next = fgetc(fp_in);
		if( c == 0x81 )
		{
			if( next == 0x41 || next == 0x43 ) return 1;
			if( next == 0x42 || next == 0x44 ) return 2;
		}
		fputc( c, fp_out );
		fputc( next, fp_out );
		return 0;
	}

	if( ('A' <= c && c <= 'Z') || ('a' <= c && c <= 'z') || ('0' <= c && c <= '9') )
		after_alphabet = 1;
	else if( c == ' ' || c == '\t' || c == '\n' )
		;
	else after_alphabet = 0;

	fputc( c, fp_out );
	return 0;
}

int is_next_alphabet( FILE* fp_in )
{
	long l = ftell( fp_in );
	int c;
	while( (c = fgetc(fp_in)) == ' ' || c == '\n' || c == '\t' )
		;
	fseek( fp_in, l, SEEK_SET );
	if( ('A' <= c && c <= 'Z') || ('a' <= c && c <= 'z') || ('0' <= c && c <= '9') )	return 1;
	else return 0;
}

/*----------------------------------------------------------------------*/
int check_replace_char( char c, FILE* fp )
{
	/* u̕KvF 2 Ԃ
	@ u̕Kv́Aʏŋ֎~ĂF1Ԃ
	@ u̕KvȂF 0 Ԃ */
	
	static int after_alphabet = 0;

	int l = ftell(fp_in);
	int ret = 0;
	if( c == ',' )
	{
		if( !(after_alphabet && is_next_alphabet(fp)) ) ret = 2;
		else ret = 1;
	}
	if( c == '.' )
	{
		if( !(after_alphabet && is_next_alphabet(fp)) ) ret = 2;
		else ret = 1;
	}
	if( c == 0x81 )
	{
		int next = fgetc(fp_in);
		if( next == 0x41 || next == 0x43
			|| next == 0x42 || next == 0x44 ) ret = 1;
		fseek( fp_in, l, SEEK_SET );
	}
	if( ('A' <= c && c <= 'Z') || ('a' <= c && c <= 'z') || ('0' <= c && c <= '9') )
		after_alphabet = 1;
	else if( c == ' ' || c == '\t' || c == '\n' )
		;
	else after_alphabet = 0;
	return ret;
}

/**********************************************************************
***                   gp@EIvV͕                 ***
**********************************************************************/
void print_usage( void )
{
	fprintf( stderr, "Usage:\n" );
	fprintf( stderr, "adjpunct [-h] [-[1-4]] filename [output-file]\n" );
	fprintf( stderr, "-h: show this message\n" );
	fprintf( stderr, "-1: use C and B(default) \n" );
	fprintf( stderr, "-2: use C and D\n" );
	fprintf( stderr, "-3: use A and B\n" );
	fprintf( stderr, "-4: use A and D\n" );
	return;
}	
/*-------------------------------------------------------------------*/
void command_line_info( int argc, _TCHAR* argv[] )
{
	static int filename_read = 0;
	for( int i = 1; i < argc; i++ )
	{
		if( argv[i][0] != '-' )
		{
			if( filename_read >= 2 )
			{
				puts( "Too many file name specified." );
				print_usage();
				exit( 1 );
			}
			else if( filename_read == 0 )
			{
				strcpy( filename, argv[i] );
				fprintf( stderr, "input from %s\n", filename );
				filename_read++;
			}
			else
			{
				strcpy( output, argv[i] );
				fprintf( stderr, "output to %s\n", output );
				filename_read++;
			}
		}
		else
		{
			switch( argv[i][1] )
			{
				case '\0':
					fprintf( stderr, "unspecified option!" );
					break;
				case '1':
					fprintf( stderr, "use C and B\n" );
					trans_mode = 1;
					break;
				case '2':
					fprintf( stderr, "use C and D\n" );
					trans_mode = 2;
					break;
				case '3':
					fprintf( stderr, "use A and B\n" );
					trans_mode = 3;
					break;
				case '4':
					fprintf( stderr, "use A and D\n" );
					trans_mode = 4;
					break;
				case 'h':
					print_usage();
					break;
				default:
					fprintf( stderr, "unknown option -%c", argv[i][1] );
					break;
			}
		}
	}
	return;
}
